// Session.cpp: implementation of the Session class.
//
//////////////////////////////////////////////////////////////////////

#include "Session.h"
#include "stdio.h"

Session*        g_pSession=NULL;         //Session object

//Thread local storage. One for each thread running.
__declspec(thread) JSObject*  g_pJSSessionObject=NULL; //Pointer to the object returned by getSession()


/* Request object definition */
static void 
MWRequestObjectDestructor(JSContext* /*pContext*/, JSObject* /*pObj*/)
{
    //printf("%ld Destroyed MWRequestObject %ld\n", GetCurrentThreadId(), pObj);
} 


/* -----------------------------------------------------*/
/* MWSession definition */
static void
MWSessionObjectDestructor(JSContext* /*pContext*/, JSObject* /*pObj*/)
{
    //Get the privates for this object.
    //printf("%ld Destroyed MWSessionObject %ld\n", GetCurrentThreadId(), pObj);
}

static JSClass MWSessionClass = {
    "MWSession", JSCLASS_HAS_PRIVATE,        //Blank for name. Cannot be created with "new"
    JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,
    JS_EnumerateStub,JS_ResolveStub,JS_ConvertStub, MWSessionObjectDestructor, JSCLASS_NO_OPTIONAL_MEMBERS
};


/* ------- Called for any method on an Session Obect by javascript land */
static JSBool 
MWSessionMethod(JSContext* pContext, JSObject* pObj, uintN argc, jsval* argv, jsval* rval)
{
    JSString* pszFunctionId=NULL;
	JSFunction* pFunction=NULL;
	char* pszFunctionName =NULL;
    *rval = JSVAL_VOID;
    JSString* pString=NULL;

    //Get the privates for this object.
    Session* pSession = (Session*)JS_GetPrivate(pContext, pObj);
    if(!pSession)
        return JS_FALSE;

	pFunction = JS_ValueToFunction(pContext, argv[-2]);
	if(pFunction)
	{
		pszFunctionId = JS_GetFunctionId(pFunction);
        if(pszFunctionId)
        {
    		pszFunctionName = JS_GetStringBytes(pszFunctionId);
            if(pszFunctionName)
            {
                if(0 == stricmp("getAttribute", pszFunctionName))
                {
                    if(argc < 1
                    || (JSTYPE_STRING != JS_TypeOfValue(pContext, argv[0]))
                    || (NULL == (pString = JS_ValueToString(pContext, argv[0]))))
                    {
                        RTS_ASSERT(FALSE);  //JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWSESSIONMSG_METHOD_REQUIRES_STRING_FOR_PARAM_1, pszFunctionName);
                        return JS_FALSE;
                    }
                    pSession->GetAttribute(pContext, std::string(JS_GetStringBytes(pString)), rval);
                }      
                else if(0 == stricmp("setAttribute", pszFunctionName))
                {
                    if(argc < 1
                    || (JSTYPE_STRING != JS_TypeOfValue(pContext, argv[0]))
                    || (NULL == (pString = JS_ValueToString(pContext, argv[0]))))
                    {
                        RTS_ASSERT(FALSE);  //JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWSESSIONMSG_METHOD_REQUIRES_STRING_FOR_PARAM_1, pszFunctionName);
                        return JS_FALSE;
                    }
                    
					if(argc < 2)
					{
                        RTS_ASSERT(FALSE);  //JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWSESSIONMSG_METHOD_REQUIRES_PARAM_2, pszFunctionName);
                        return JS_FALSE;
                    }
					pSession->SetAttribute(pContext, std::string(JS_GetStringBytes(pString)), argv[1]);

				/*	JSString* pValue=NULL;
                    if(argc < 2
                    || (NULL == (pValue = JS_ValueToString(pContext, argv[1]))))
                    {
                        JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWSESSIONMSG_METHOD_REQUIRES_STRING_FOR_PARAM_2, pszFunctionName);
                        return JS_FALSE;
                    }
                    //char* pszBytes = JS_GetStringBytes(pValue);
                    //pSession->SetAttribute(std::string(JS_GetStringBytes(pString)), std::string(pszBytes));

                    pSession->SetAttribute(pContext, std::string(JS_GetStringBytes(pString)), &argv[1]);
		        */
				}    
                else if(0 == stricmp("getAttributeNames", pszFunctionName))
                {
                    JSObject* pArray=NULL;
                    JSString* pString=NULL;
                    jsval oElement;
                    
                    SESSION_ATTRIBUTE_MAP* pAttributeMap = pSession->GetAttributeMap();

                    int iNumAttributes = pAttributeMap->size();
                    if(iNumAttributes)
                    {
                        jsint  iLen=0;
                        pArray = JS_NewArrayObject(pContext, 0, NULL); //create it empty.
                
                        if(pArray)
                        {
                            StringMap::iterator oIter;
                            SESSION_ATTRIBUTE_MAP::iterator oAttributeIter;
                            oAttributeIter = pAttributeMap->begin();
                            while(oAttributeIter != pAttributeMap->end())
                            {
                               pString = JS_NewStringCopyZ(pContext, (*oAttributeIter).first.c_str());
                               oElement = STRING_TO_JSVAL(pString);
                               JS_SetElement(pContext, pArray, iLen, &oElement);
                               ++oAttributeIter;
                               iLen++;
                            }
                        }

                        *rval = OBJECT_TO_JSVAL(pArray);
                    }
                    else
                    {
                        *rval = JSVAL_NULL;
                    }
                }      
                else if(0 == stricmp("removeAttribute", pszFunctionName))
                {
                    if(argc < 1
                    || (JSTYPE_STRING != JS_TypeOfValue(pContext, argv[0]))
                    || (NULL == (pString = JS_ValueToString(pContext, argv[0]))))
                    {
                        RTS_ASSERT(false); //JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWSESSIONMSG_METHOD_REQUIRES_STRING_FOR_PARAM_1, pszFunctionName);
                        return JS_FALSE;
                    }
                    pSession->RemoveAttribute(std::string(JS_GetStringBytes(pString)));
                }    
            }
        }
    }
    return JS_TRUE;
}

 static JSFunctionSpec MWSession_Methods[] = {
       {"getAttribute",                 MWSessionMethod, 1, 0, 0},
       {"getAttributeNames",            MWSessionMethod, 0, 0, 0},
       {"removeAttribute",              MWSessionMethod, 1, 0, 0}, 
       {"setAttribute",                 MWSessionMethod, 2, 0, 0},
       {0,0,0,0,0},	//Terminates the table of functions. REQUIRED.
   };


static JSBool
MWRequestMethod(JSContext* pContext, JSObject* /*obj*/, uintN argc, jsval* argv, jsval* rval)
{
     /*
     * Look in argv for argc actual parameters, set *rval to return a
     * value to the caller.
     */
    JSString* pszFunctionId=NULL;
	JSFunction* pFunction=NULL;
	char* pszFunctionName =NULL;
    //JSString* pString=NULL;
    *rval = JSVAL_VOID;
	JSObject* pGlobal = JS_GetGlobalObject(pContext);

	pFunction = JS_ValueToFunction(pContext, argv[-2]);
	if(pFunction)
	{
		pszFunctionId = JS_GetFunctionId(pFunction);
        if(pszFunctionId)
        {
    		pszFunctionName = JS_GetStringBytes(pszFunctionId);
            if(pszFunctionName)
            {
                if(0 == stricmp("getsession", pszFunctionName))
                {
	                JSBool bCreateNew=0;
	                JSBool bAutoUpdateLastAccessTime=JS_TRUE;
					bool bWasAutoUpdateFlagPassed=false;

                    if(argc < 1 || !JS_ValueToBoolean(pContext, argv[0], &bCreateNew))
                    {
                        RTS_ASSERT(false);//JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWREQUESTMSG_METHOD_REQUIRES_BOOLEAN_FOR_PARAM_1, pszFunctionName);
                        return JS_FALSE;
                    }

                    if(argc > 1)
					{
						if(!JS_ValueToBoolean(pContext, argv[1], &bAutoUpdateLastAccessTime))
						{
	                        RTS_ASSERT(false);//JS_ReportErrorNumber(pContext, MW_GetErrorMessage, NULL, MWREQUESTMSG_METHOD_REQUIRES_BOOLEAN_FOR_PARAM_2, pszFunctionName);
		                    return JS_FALSE;
						}
						else
						{
							bWasAutoUpdateFlagPassed=true;
						}
                    }
                    
                    //If the javascript object doesn't exist yet create it now.
                    //Will happen for a new session and existing session that needs a new object in javascript land.
					//Also if they just invalidated a session and want to create a new one then release our
					//javascript land object and create another one.
                    if(NULL == g_pJSSessionObject)
                    {
                        g_pJSSessionObject = JS_DefineObject(pContext, pGlobal, "__SessionObject__", &MWSessionClass, NULL /*pSessionClass*/, JSPROP_READONLY|JSPROP_PERMANENT);
                        if(!g_pJSSessionObject)
                            return JS_FALSE;

                        if(!JS_DefineFunctions(pContext, g_pJSSessionObject, MWSession_Methods))
                            return JS_FALSE;

                        //Alias to upper case verions for the Microsoft folks out there.
                        JSBool bOK=JS_FALSE;
                        bOK = JS_AliasProperty(pContext, g_pJSSessionObject, "getAttribute",           "GetAttribute");
                        bOK = JS_AliasProperty(pContext, g_pJSSessionObject, "getAttributeNames",      "GetAttributeNames");
                        bOK = JS_AliasProperty(pContext, g_pJSSessionObject, "removeAttribute",        "RemoveAttribute");
                        bOK = JS_AliasProperty(pContext, g_pJSSessionObject, "setAttribute",           "SetAttribute");

                        JS_SetPrivate(pContext, g_pJSSessionObject, g_pSession);
                        *rval = OBJECT_TO_JSVAL(g_pJSSessionObject);
                        return JS_TRUE;
                    }
                    else
                    {
                        //Otherwise return the already created ojbect
                        //printf("%d returned existing Session object %ld\n", GetCurrentThreadId(), g_pJSSessionObject);
                        *rval = OBJECT_TO_JSVAL(g_pJSSessionObject);
                        return JS_TRUE;
                    }
                }
                else if(0 == stricmp("end", pszFunctionName))
                {
                    //STOP PROCESSING javascript and terminate.
                    return JS_FALSE;
                }
				else
				{
					*rval = JSVAL_NULL;
				}
            }
        }
    }
    
    return JS_TRUE;
}


/* ------- Page Object (or global) */

//Add the following properites to the MWFile Class
enum MWGlobal_tinyid
{
    MWGLOBAL_PAGE
};

static JSBool
MWGlobal_GetProperty(JSContext* cx, JSObject* /*obj*/, jsval id, jsval *vp)
{
	*vp = JSVAL_VOID;   //Default to void.
    jsint   iPropId = JSVAL_TO_INT(id);

	//If they want the page object just return the global object itself.
    if(iPropId == MWGLOBAL_PAGE)
    {
       *vp = OBJECT_TO_JSVAL(JS_GetGlobalObject(cx));
	}

    return JS_TRUE;
}

static JSPropertySpec global_props[] = {
   {"page",          MWGLOBAL_PAGE,       JSPROP_ENUMERATE | JSPROP_READONLY, MWGlobal_GetProperty, NULL },
   {0}
};    


void
InitMWScriptingObjects(JSContext* pContext, JSObject* pGlobal)
{
    //Now try to create custom objects for use in the scripts
    static JSClass MWRequestClass = {
        "Request", JSCLASS_HAS_PRIVATE,
        JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,
        JS_EnumerateStub,JS_ResolveStub,JS_ConvertStub, MWRequestObjectDestructor, JSCLASS_NO_OPTIONAL_MEMBERS
    };

	//Test for adding functions into the scripting language.
     static JSFunctionSpec        MWRequest_Methods[] = {
            {"getSession",        MWRequestMethod, 1, 0, 0},
            {"end",               MWRequestMethod, 0, 0, 0},      //Used to stop script processing.
           {0,0,0,0,0},	//Terminates the table of functions. REQUIRED.
       };

    JSObject* pRequestObject = JS_DefineObject(pContext, pGlobal, "request", &MWRequestClass, NULL /*pRequestClass*/, JSPROP_READONLY | JSPROP_PERMANENT);
    if (!pRequestObject)
        return;
 
    if (!JS_DefineFunctions(pContext, pRequestObject, MWRequest_Methods))
        return;

    //Alias as upper case versions for ease of use for Microsoft ASP folks.
    JS_AliasProperty(pContext, pRequestObject, "getSession",        "GetSession");
    JS_AliasProperty(pContext, pRequestObject, "end",               "End");   

	//Add some properties to the global object.
	JSBool bDefineOK;
	bDefineOK = JS_DefineProperties(pContext, pGlobal, global_props);
	bDefineOK = JS_AliasProperty(pContext, pGlobal, "page", "Page");
}




SessionAttrValue::SessionAttrValue(JSType iType)
{
	m_iType = iType;
	m_pValue = NULL;
}

SessionAttrValue::~SessionAttrValue()
{
	if(m_pValue)
	{
		switch(m_iType)
		{
			//Fall through for these types and do nothing.
			case JSTYPE_VOID:
			case JSTYPE_LIMIT:
			case JSTYPE_OBJECT:
			case JSTYPE_FUNCTION:
				break;

			case JSTYPE_STRING:
				delete (std::string*)m_pValue;
				break;

			case JSTYPE_NUMBER:
				delete (double*)m_pValue;
				break;

			case JSTYPE_BOOLEAN:
				delete (bool*)m_pValue;
				break;
		}
	}
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Session::Session(JSRuntime* pRuntime)
{
	m_pJSRuntime = pRuntime;
    InitializeCriticalSection(&m_oSessionCritSect);
}


Session::~Session()
{
	ClearAllAttributes();

    DeleteCriticalSection(&m_oSessionCritSect);
}

void Session::ClearAllAttributes()
{
	//Iterate all the session attributes and clean them up..
    EnterCriticalSection(&m_oSessionCritSect);

    SESSION_ATTRIBUTE_MAP::iterator itAttrib;
    itAttrib = m_oAttributeMap.begin();
    SessionAttrValue* pAttrValue=NULL;

	while(itAttrib != m_oAttributeMap.end())
    {
		pAttrValue = (*itAttrib).second;
		if(pAttrValue)
		{
            switch(pAttrValue->m_iType)
            {
                case JSTYPE_OBJECT:
                case JSTYPE_FUNCTION:
				    if(!JS_RemoveRootRT(m_pJSRuntime, pAttrValue->m_pValue))
				    {
					    RTS_ASSERT(false);
				    }
                    JS_free(NULL, (jsval*)pAttrValue->m_pValue);
                    break;

			    case JSTYPE_STRING:
                    delete (std::string*)pAttrValue->m_pValue;
                    break;

                case JSTYPE_NUMBER:
	                delete (double*)pAttrValue->m_pValue;
    			    break;
    		    
                case JSTYPE_BOOLEAN:
                    delete (bool*)pAttrValue->m_pValue;
                    break;
           }
           delete pAttrValue;
           pAttrValue=NULL;
        }
        ++itAttrib;
	}

    LeaveCriticalSection(&m_oSessionCritSect);

    m_oAttributeMap.clear();
}

void  Session::IncrementIsUseCount()
{
    EnterCriticalSection(&m_oSessionCritSect);
    m_liInUseCount++;
//    printf("%d Incrementing Session Ref Count= %d\n", GetCurrentThreadId(), m_liInUseCount);
    LeaveCriticalSection(&m_oSessionCritSect);
}

void  Session::DecrementIsUseCount()
{
    EnterCriticalSection(&m_oSessionCritSect);
    m_liInUseCount--;
//    printf("%d Decrementing Session Ref Count= %d\n", GetCurrentThreadId(), m_liInUseCount);
    LeaveCriticalSection(&m_oSessionCritSect);
}

long  Session::GetInUseCount()
{
    long liUseCount=0;

    EnterCriticalSection(&m_oSessionCritSect);
    liUseCount = m_liInUseCount;
    LeaveCriticalSection(&m_oSessionCritSect);
    
    return liUseCount;
}


bool Session::SetAttribute(JSContext* pContext, std::string sName, jsval oJSValue)
{
    //printf("*********** Calling SetAttribute(%s) **************\n", sName.c_str());

    //Just overwrite it value if it already exists.
    EnterCriticalSection(&m_oSessionCritSect);
  
	//Check if they are overwritting and existing value...
	SessionAttrValue* pExistingValue = NULL;
	SESSION_ATTRIBUTE_MAP::iterator oFind = m_oAttributeMap.find(sName);
  	if(oFind != m_oAttributeMap.end())
		pExistingValue = oFind->second;
  
	JSType iType = JS_TypeOfValue(pContext, oJSValue);
	
	//If we have an existing value of the same name bug different type
	//then just remove it.  Otherwise overwrite the existing value.
	if(pExistingValue && pExistingValue->m_iType != iType)
	{
		RemoveAttribute(sName);
		pExistingValue=NULL;
	}

	SessionAttrValue* pNewValue = NULL;
	if(!pExistingValue)
		pNewValue = new SessionAttrValue(iType);
	
	JSString* pString=NULL;
	jsdouble dJDouble=NULL;
	switch(iType)
	{
		//Fall through for these types and do nothing.
		case JSTYPE_VOID:
		case JSTYPE_LIMIT:
			break;	//Do nothing for void. Just store its type.

		case JSTYPE_OBJECT:
		case JSTYPE_FUNCTION:
			//If we already have the value no point in rooting it twice!!
			if(!pExistingValue)
			{
				pNewValue->m_pValue = (jsval*) JS_malloc(pContext, sizeof(jsval));
				//pNewValue->m_pValue = (void*)new jsval;
				*((jsval*)pNewValue->m_pValue) = oJSValue;

				if(JSVAL_IS_GCTHING(oJSValue))
				{
					//Now SEAL the object to make it threadsafe for multiple access form one or more threads.
					//JS_SealObject(pContext, JSVAL_TO_OBJECT(oJSValue), JS_FALSE);	//Do NOT seal child objects. Will lock the global!

					if(!JS_AddNamedRootRT(m_pJSRuntime, pNewValue->m_pValue, "SessionAttributeValue"))
					{
						RTS_ASSERT(false);
					}
				}
				else
				{
					RTS_ASSERT(false);
				}
			}
			break;

		case JSTYPE_STRING:
			pString = JS_ValueToString(pContext, oJSValue);
			if(pString)
			{
				//Overwrite the existing value if we have one.
				if(pExistingValue)
					*((std::string*)(pExistingValue->m_pValue)) = JS_GetStringBytes(pString);
				else
					pNewValue->m_pValue = (void*)new std::string(JS_GetStringBytes(pString));
			}
			break;

		case JSTYPE_NUMBER:
			if(!JS_ValueToNumber(pContext, oJSValue, &dJDouble))
			{
				RTS_ASSERT(false);
			}

			//Overwrite the existing value if we have one.
			if(pExistingValue)
			{
				*((double*)(pExistingValue->m_pValue)) = dJDouble;
			}
			else
			{
				pNewValue->m_pValue = new double();
				*((double*)(pNewValue->m_pValue)) = dJDouble;
			}
			break;

		case JSTYPE_BOOLEAN:
			JSBool bValue;
			if(!JS_ValueToBoolean(pContext, oJSValue, &bValue))
			{
				RTS_ASSERT(false);
			}

			//Overwrite the existing value if we have one.
			if(pExistingValue)
			{
				*((bool*)(pExistingValue->m_pValue)) = (JS_FALSE != bValue);
			}
			else
			{
				pNewValue->m_pValue = new bool();
				*((bool*)(pNewValue->m_pValue)) = (JS_FALSE != bValue);
			}
			break;
	}

	if(NULL == pExistingValue)
		m_oAttributeMap.insert(SESSION_ATTRIBUTE_MAP::value_type(sName, pNewValue));
    
    LeaveCriticalSection(&m_oSessionCritSect);
	return true;
}

void  Session::RemoveAttribute(std::string sName)
{
    EnterCriticalSection(&m_oSessionCritSect);

    SESSION_ATTRIBUTE_MAP::iterator oFind;
    oFind = m_oAttributeMap.find(sName);
    if(oFind != m_oAttributeMap.end())
    {
		SessionAttrValue* pAttrValue = (*oFind).second;
		if(pAttrValue)
		{
            switch(pAttrValue->m_iType)
            {
                case JSTYPE_OBJECT:
                case JSTYPE_FUNCTION:
				    if(!JS_RemoveRootRT(m_pJSRuntime, pAttrValue->m_pValue))
				    {
					    RTS_ASSERT(false);
				    }
                    JS_free(NULL, (jsval*)pAttrValue->m_pValue);
                    break;

			    case JSTYPE_STRING:
                    delete (std::string*)pAttrValue->m_pValue;
                    break;

                case JSTYPE_NUMBER:
	                delete (double*)pAttrValue->m_pValue;
    			    break;
    		    
                case JSTYPE_BOOLEAN:
                    delete (bool*)pAttrValue->m_pValue;
                    break;
           }
            pAttrValue->m_pValue = NULL;	//Null it so somebody else doens't try and delete it later.
		}
		delete pAttrValue;
        pAttrValue=NULL;
		m_oAttributeMap.erase(oFind);
	}

    LeaveCriticalSection(&m_oSessionCritSect);
}


bool
Session::GetAttribute(JSContext* pContext, std::string sName, jsval* pJSValue)
{
//	bool bReturn=true;
    //printf("*********** Calling GetAttribute(%s) **************\n", sName.c_str());

    EnterCriticalSection(&m_oSessionCritSect);

	*pJSValue = JSVAL_NULL;
	SessionAttrValue* pAttrValue=NULL;

    SESSION_ATTRIBUTE_MAP::iterator oFind;
    oFind = m_oAttributeMap.find(sName);
    if(oFind != m_oAttributeMap.end())
 		pAttrValue = (*oFind).second;
	else
	{
		LeaveCriticalSection(&m_oSessionCritSect);
		return false;
	}

	switch(pAttrValue->m_iType)
	{
		case JSTYPE_VOID:
			*pJSValue = JSVAL_VOID;
			break;
		case JSTYPE_LIMIT:
			*pJSValue = JSVAL_NULL;
			break;

		case JSTYPE_OBJECT:
		case JSTYPE_FUNCTION:
			*pJSValue = *((jsval*)pAttrValue->m_pValue);
			break;

		case JSTYPE_STRING:
			*pJSValue = STRING_TO_JSVAL(JS_NewStringCopyZ(pContext, ((std::string*)pAttrValue->m_pValue)->c_str()));
			break;
		case JSTYPE_NUMBER:
			*pJSValue = DOUBLE_TO_JSVAL(JS_NewDouble(pContext, *((double*)pAttrValue->m_pValue)));
			break;

		case JSTYPE_BOOLEAN:
			*pJSValue = BOOLEAN_TO_JSVAL(*((bool*)pAttrValue->m_pValue));
			break;
    }

    LeaveCriticalSection(&m_oSessionCritSect);
    return false;
};

