// Session.h: interface for the Session class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(_SESSION_H__INCLUDED_)
#define _SESSION_H__INCLUDED_

#include <windows.h>
#include <string>
#include <map>
#include "jsapi.h"	//jsval definition


typedef std::map<std::string, std::string>	StringMap;
typedef StringMap::iterator				itrStringMap;
typedef StringMap::reverse_iterator		revItrStringMap;

#pragma warning( disable : 4786)  // Disable warning "identifier was truncated to '255'"
#pragma warning( disable : 4503)  // Disable warning "decorated name length exceeded, name was truncated...


#if defined(_DEBUG)
#define RTS_ASSERT(_expr) \
    ((_expr)?((void)0):DebugBreak())
#else
	#define RTS_ASSERT(expr) ((void) 0)
#endif

//This class is a simple holder for atributes that are saved in the session.
//Attributes can be of various types including
//numeric, string, or objects.
class SessionAttrValue  
{
public:
	
	SessionAttrValue(JSType iType);
	virtual	~SessionAttrValue();

	JSType			m_iType;	//Holds one of the following types:
								// JSTYPE_VOID
								// JSTYPE_OBJECT
								// JSTYPE_FUNCTION
								// JSTYPE_STRING
								// JSTYPE_NUMBER
								// JSTYPE_BOOLEAN
								// JSTYPE_NULL
	
	//If the value is of type OBJECT or FUNCTION the jsval pointer is rooted in the runtime so it won't get garbage collected.

	void*			m_pValue;	//Will be one of the following values:
								// std::string
								// jsval	
								// double
								// bool
};


//Name value pairs of attributes associated with this session.
//typedef std::map<std::string, std::string> SESSION_ATTRIBUTE_MAP;

typedef std::map<std::string, SessionAttrValue*> SESSION_ATTRIBUTE_MAP;

extern void InitMWScriptingObjects(JSContext* pContext, JSObject* pGlobal);


class Session
{
protected:

private:
    //Pointer to spidermonkey runtime so that we can "root" jsval objects in the runtime so as to be
	//able to reuse them across any contexts in the pool for session attributes.
	JSRuntime*				m_pJSRuntime;

	SESSION_ATTRIBUTE_MAP   m_oAttributeMap;
 
    //Contains a reference count of the number of threads that are using this object
    //Once the ref count becomes zero and m_bInvalidated is true then it is safe to delete
    //this object.  Its possible that several simultaneous threads might be accessing this object.
    //Its a weird scenario but it CAN happen. Normally only if somebody refreshes the browser
    //many times really fast before the first request is completed.
	//Can also happen if we are processing XMLHttpRequests
    long                    m_liInUseCount;     
    
    CRITICAL_SECTION        m_oSessionCritSect;

public:

    Session(JSRuntime* pRuntime);
	virtual ~Session();

    bool         SetAttribute(JSContext* pContext, std::string sName, jsval oJSValue);
    void         RemoveAttribute(std::string sName);
    bool         GetAttribute(JSContext* pContext, std::string sName, jsval* pJSValue);

    SESSION_ATTRIBUTE_MAP*  GetAttributeMap()				{return &m_oAttributeMap;}
	void		 ClearAllAttributes();

    void         IncrementIsUseCount();
    void         DecrementIsUseCount();
    long         GetInUseCount();
};

extern Session* g_pSession;

#endif // !defined(_SESSION_H__INCLUDED_)
